/*
 * Copyright 2009-2018 Joshua Bronson. All Rights Reserved.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

h1, h2, h3, h4, h5 { font-weight: bold !important; }

ul li { margin-bottom: 8px; }

/* Float the logo to the right of the status badges on the homepage */
#bidict img[alt="bidict logo"] { float: right; }

/* Can remove this when https://github.com/bitprophet/alabaster/pull/125 lands */
a.image-reference:hover { border-bottom: 0 !important; }

/* override the inherited "max-width: 100%" in case a badge fails to load */
#status img { max-width: inherit; }

.sphinxsidebar h3 { margin-top: 20px !important; }
.sphinxsidebar h3.donation { margin-top: 50px !important; }
